# -*- coding: utf-8 -*-

from wow_wtf.api import exp03_wotlk
from wow_wtf.logger import logger


def test():
    from wow_wtf.tests.exp03_wotlk.acc_dataset import dir_here, ds

    path = dir_here.joinpath("acc_enum.py")
    content = ds.to_module(
        import_line="from wow_wtf.tests.exp03_wotlk.acc_dataset import ds",
    )
    path.write_text(content)

    content = exp03_wotlk.to_module(
        dir_here,
        import_dir_root_line="from wow_wtf.tests.exp03_wotlk.wtf_dataset import dir_root",
    )
    dir_here.joinpath("wtf_enum.py").write_text(content, encoding="utf-8")

    from wow_wtf.tests.exp03_wotlk.wtf_mapping import wtf_mapping

    wtf_mapping.client.dir_wtf.remove_if_exists()
    real_run = True

    with logger.disabled(
        disable=True,  # DON't show log
        # disable=False, # show log
    ):
        wtf_mapping.apply_client_config(real_run)

        wtf_mapping.apply_account_user_interface(real_run)
        wtf_mapping.apply_account_saved_variables(real_run)

        wtf_mapping.apply_character_user_interface(real_run)
        wtf_mapping.apply_character_chat(real_run)
        wtf_mapping.apply_character_keybinding(real_run)
        wtf_mapping.apply_character_layout(real_run)
        wtf_mapping.apply_character_addons(real_run)
        wtf_mapping.apply_character_saved_variables(real_run)


if __name__ == "__main__":
    from wow_wtf.tests import run_cov_test

    run_cov_test(__file__, "wow_wtf.exp03_wotlk", preview=False)



# -*- coding: utf-8 -*-

"""
这个模块可以生成一个对所有的 WTF config 文件模板的枚举 Python 模块.
详情请看 :func:`to_module` 函数的文档.
"""

import dataclasses
from pathlib_mate import Path


@dataclasses.dataclass
class Folder:
    """
    代表着一类 WTF 配置的模板集合的文件夹. 例如 ``11_account_user_interface`` 就是
    所有账号的用户界面配置文件的文件夹.

    :param name: 文件夹的名字, 例如 ``11_account_user_interface``
    :param classname: 生成的 Python 枚举类的名字, 例如它是 ``AccountUserInterface``
        的话, 最终生成的 Python 枚举类就是 ``class AccountUserInterfaceEnum:``
    :param file_ext: 模板文件的扩展名, 例如 ``.txt``, 我们会用这个来筛选出被视为 "模板" 的文件.
    """

    name: str = dataclasses.field()
    classname: str = dataclasses.field()
    file_ext: str = dataclasses.field()


folder_list = [
    Folder(
        name="01_client_config",
        classname="ClientConfig",
        file_ext=".txt",
    ),
    Folder(
        name="11_account_user_interface",
        classname="AccountUserInterface",
        file_ext=".txt",
    ),
    Folder(
        name="12_account_macros",
        classname="AccountMacros",
        file_ext=".txt",
    ),
    Folder(
        name="13_account_saved_variables",
        classname="AccountSavedVariables",
        file_ext=".lua",
    ),
    Folder(
        name="21_character_user_interface",
        classname="CharacterUserInterface",
        file_ext=".txt",
    ),
    Folder(
        name="22_character_chat",
        classname="CharacterChat",
        file_ext=".txt",
    ),
    Folder(
        name="23_character_keybindings",
        classname="CharacterKeybindings",
        file_ext=".txt",
    ),
    Folder(
        name="24_character_layout",
        classname="CharacterLayout",
        file_ext=".txt",
    ),
    Folder(
        name="25_character_addons",
        classname="CharacterAddons",
        file_ext=".txt",
    ),
    Folder(
        name="26_character_macros",
        classname="CharacterMacros",
        file_ext=".txt",
    ),
    Folder(
        name="27_character_saved_variables",
        classname="CharacterSavedVariables",
        file_ext=".lua",
    ),
]
"""
列出了所有 WTF 配置模板文件夹的规则.
"""


def slugify(s: str) -> str:
    """
    将字符串转换成一个合法的 Python 变量名.
    """
    return s.replace(" ", "_").replace("-", "_").replace("/", "__").replace("\\", "__")


def get_var_name(
    dir: Path,
    path: Path,
):
    """
    从文件路径生成一个合法的 Python 变量名作为 Enum 枚举值的变量名. ``path`` 是这个模板文件
    的路径, 而 ``dir`` 则是 :class:`Folder` 文件夹的路径.
    例如我们有一个 ``${HOME}/11_account_user_interface/default.txt`` 文件, 而 ``dir`` 是
    ``${HOME}/11_account_user_interface``. 那么这个模板文件的变量名就会是 ``default``.

    :param dir:
    :param path:
    """
    relpath = path.relative_to(dir)
    var_name = slugify(str(relpath)).split(".")[0]
    if var_name[0].isalpha() is False:  # 如果第一个字符不是字母, 那么加上一个 f_ (file)
        var_name = "f_" + var_name
    return var_name


def to_module(
    dir_root: Path,
    import_dir_root_line: str,
) -> str:
    """
    根据前面 ``folder_list`` 中的设定, 扫描 ``dir_root`` 文件夹下的所有文件, 生成一个
    Python 模块的字符串. 这个模块包含了所有的 WTF 模板文件的枚举.

    例如你有这样的文件结构:

    .. code-block::

        /{dir_root}/01_client_config
            ...
        /{dir_root}/11_account_user_interface
            default.txt
            ...
        /{dir_root}/12_account_macros
            ...
        /{dir_root}/13_account_saved_variables
            ...
        /{dir_root}/21_character_user_interface
            default.txt
            ...
        /{dir_root}/22_character_chat
            ...
        /{dir_root}/23_character_keybindings
            ...
        /{dir_root}/24_character_layout
            ...
        /{dir_root}/25_character_addons
            ...
        /{dir_root}/26_character_macros
            ...
        /{dir_root}/27_character_saved_variables
            ...

    那么最终生成的 Python 模块请参考 :ref:`generate-wtf-config-enum-module`

    :param dir_root: WTF 配置模板文件的根目录.
    :param import_dir_root_line: 这一行要导入一个 dir_root 对象, 也就是我们扫描的
        WTF 配置模板文件的根目录. 用于里面的 enum 中的路径的拼接.
    """
    lines = [
        "# -*- coding: utf-8 -*-",
        "",
        f"{import_dir_root_line}",
        "# fmt: off",
        "",
    ]
    tab = " " * 4
    for folder in folder_list:
        dir_folder = dir_root / folder.name
        lines.append(f"class {folder.classname}Enum:")
        paths = list(dir_folder.select_by_ext(folder.file_ext, recursive=True))
        paths.sort()
        if len(paths):
            for path in paths:
                var_name = get_var_name(dir_folder, path)
                relpath = path.relative_to(dir_root)
                joinpath_arg = ", ".join([f'"{part}"' for part in relpath.parts])
                dir_root.joinpath("{}")
                lines.append(
                    f"{tab}{var_name} = dir_root.joinpath({joinpath_arg}) # file://{path}"
                )
        else:
            lines.append(f"{tab}pass")
        lines.append("")
        lines.append("")
    lines.append("# fmt: on")
    lines.append("")
    return "\n".join(lines)



# -*- coding: utf-8 -*-

"""
这个模块的可以让你对 Account / Character 和 WTF 配置文件的模板进行排列组合, 然后一键将你的
配置文件应用到你的客户端中的 WTF 目录下.
"""

import typing as T
import dataclasses
from pathlib_mate import Path
from functools import cached_property

from jinja2 import Template
from wow_acc.api import Account, Character

from ..logger import logger
from ..utils import group_by


@dataclasses.dataclass
class Client:
    r"""
    代表着一个具体魔兽世界客户端. 你必须要制定这个客户端的目录. 然后就可以用各种 method 来
    获取对应的 WTF 配置文件的绝对路径了.

    :param dir: 客户端目录, 根据此目录可以定位其他的目录. 例如 "C:\Program Files\World of Warcraft"
    :param locale: 客户端语种, 例如 enUS, zhCN, zhTW 等.
    """

    dir: Path = dataclasses.field()
    locale: str = dataclasses.field()

    @property
    def dir_wtf(self) -> Path:
        return self.dir / "WTF"

    @property
    def client_config(self) -> Path:
        r"""
        This file stores client level user interface configuration.

        Example: ``C:\...\WTF\Config.wtf``
        """
        return self.dir_wtf / "Config.wtf"

    def _get_acc_dir(self, account: "Account") -> Path:
        r"""
        Example: ``C:\...\WTF\Account\MYACCOUNT``
        """
        return self.dir_wtf / "Account" / account.wtf_account_name

    def get_account_config_cache_wtf(
        self,
        account: "Account",
    ) -> Path:
        r"""
        This file stores account level user interface configuration.

        Example: ``C:\...\WTF\Account\MYACCOUNT\config-cache.wtf``
        """
        return self._get_acc_dir(account) / "config-cache.wtf"

    def get_account_bindings_cache_wtf(
        self,
        account: "Account",
    ) -> Path:  # pragma: no cover
        r"""
        This file stores account level key bindings.

        Example: ``C:\...\WTF\Account\MYACCOUNT\bindings-cache.wtf``
        """
        return self._get_acc_dir(account) / "bindings-cache.wtf"

    def get_account_macros_cache_txt(
        self,
        account: "Account",
    ) -> Path:  # pragma: no cover
        r"""
        This file stores account level macro configurations.

        Example: ``C:\...\WTF\Account\MYACCOUNT\macros-cache.txt``
        """
        return self._get_acc_dir(account) / "macros-cache.txt"

    def get_account_saved_variables(
        self,
        account: "Account",
        file: str,
    ) -> Path:
        r"""
        This file stores per AddOn account level saved variables.

        Example: ``C:\...\WTF\Account\MYACCOUNT\SavedVariables\AtlasLoot.lua``
        """
        return self._get_acc_dir(account) / "SavedVariables" / file

    def _get_char_dir(self, character: "Character") -> Path:
        r"""
        Example: ``C:\...\WTF\Account\MYACCOUNT\MyServer\Mycharacter\``
        """
        return (
            self._get_acc_dir(character.account)
            / character.realm_name
            / character.titled_character_name
        )

    def get_character_config_cache_wtf(
        self,
        character: "Character",
    ) -> Path:
        r"""
        This file stores character level user interface configuration.

        Example: ``C:\...\WTF\Account\MYACCOUNT\MyServer\Mycharacter\config-cache.wtf``
        """
        return self._get_char_dir(character) / "config-cache.wtf"

    def get_character_chat_cache_txt(
        self,
        character: "Character",
    ) -> Path:
        r"""
        This file stores character level chat cache configuration.

        Example: ``C:\...\WTF\Account\MYACCOUNT\MyServer\Mycharacter\chat-cache.txt``
        """
        return self._get_char_dir(character) / "chat-cache.txt"

    def get_character_bindings_cache_wtf(
        self,
        character: "Character",
    ) -> Path:
        r"""
        This file stores character level key bindings.

        Example: ``C:\...\WTF\Account\MYACCOUNT\MyServer\Mycharacter\bindings-cache.wtf``
        """
        return self._get_char_dir(character) / "config-cache.wtf"

    def get_character_layout_local_txt(
        self,
        character: "Character",
    ) -> Path:
        r"""
        This file stores character level UI layout configurations.

        Example: ``C:\...\WTF\Account\MYACCOUNT\MyServer\Mycharacter\layout-local.txt``
        """
        return self._get_char_dir(character) / "layout-local.txt"

    def get_character_addons_txt(
        self,
        character: "Character",
    ) -> Path:
        r"""
        This file stores character level UI AddOns enable / disable configurations.

        Example: ``C:\...\WTF\Account\MYACCOUNT\MyServer\Mycharacter\AddOns.txt``
        """
        return self._get_char_dir(character) / "AddOns.txt"

    def get_character_macros_cache_txt(
        self,
        character: "Character",
    ) -> Path:  # pragma: no cover
        r"""
        This file stores character level macro configurations.

        Example: ``C:\...\WTF\Account\MYACCOUNT\MyServer\Mycharacter\macros-cache.txt``
        """
        return self._get_char_dir(character) / "macros-cache.txt"

    def get_character_saved_variables(
        self,
        character: "Character",
        file: str,
    ) -> Path:
        r"""
        This file stores per AddOn character level saved variables.

        Example: ``C:\...\WTF\Account\MYACCOUNT\MyServer\Mycharacter\SavedVariables\tlasLoot.lua``
        """
        return self._get_char_dir(character) / "SavedVariables" / file


@dataclasses.dataclass
class BaseMapping:
    """
    Mapping 是一个 Account / Character 和一个 WTF 配置文件模板的组合.
    """

    file: Path = dataclasses.field()

    @cached_property
    def tpl(self) -> Template:
        """
        Jinja template 对象. 它会被缓存下来以便复用.
        """
        return Template(self.file.read_text(encoding="utf-8"))


@dataclasses.dataclass
class AccLvlMapping(BaseMapping):
    """
    ``Account`` 和一个 WTF 配置文件模板的组合.

    :param acc: `wow_acc.api.Account <https://wow-acc.readthedocs.io/en/latest/wow_acc/model.html#wow_acc.model.Account>`_ 对象.
    :param file: WTF 配置文件模板的绝对路径.
    """

    acc: Account = dataclasses.field()
    file: Path = dataclasses.field()

    @classmethod
    def make_many(
        cls,
        accounts: T.Iterable[Account],
        file: T.Union[Path, T.Iterable[Path]],
    ):
        """
        生成多个 ``Account`` 和 WTF 配置文件模板的组合. 让你写更少的代码.

        :param accounts: Account 对象集合.
        :param file: 单个文件或是多个文件的集合.
        """
        if isinstance(file, Path):
            return [cls(acc=acc, file=file) for acc in accounts]
        else:
            lst = list()
            for acc in accounts:
                for f in file:
                    lst.append(cls(acc=acc, file=f))
            return lst


@dataclasses.dataclass
class CharLvlMapping(BaseMapping):
    """
    ``Character`` 和一个 WTF 配置文件模板的组合.

    :param char: `wow_acc.api.Character <https://wow-acc.readthedocs.io/en/latest/wow_acc/model.html#wow_acc.model.Character>`_ 对象.
    :param file: WTF 配置文件模板的绝对路径.
    """

    char: Character = dataclasses.field()
    file: Path = dataclasses.field()

    @classmethod
    def make_many(
        cls,
        characters: T.Iterable[Character],
        file: T.Union[Path, T.List[Path]],
    ):
        """
        生成多个 ``Character`` 和 WTF 配置文件模板的组合. 让你写更少的代码.

        :param characters: Character 对象集合.
        :param file: 单个文件或是多个文件的集合.
        """
        if isinstance(file, Path):
            return [cls(char=char, file=file) for char in characters]
        else:
            lst = list()
            for char in characters:
                for f in file:
                    lst.append(cls(char=char, file=f))
            return lst


def apply(
    path: Path,
    content: str,
    real_run: bool = False,
    verbose: bool = True,
):
    r"""
    Apply a content to a file (Write to the file in ``World of Warcraft\WTF\...``).

    :param path: The file path to write to.
    :param content: The content to write.
    :param real_run: If True, do not write to the file.
    :param verbose: If True, print log.
    """
    if verbose:
        # 使用 file:// URI 使得可以点击打印在 Console 中的输出直接跳转到文件
        logger.info(f"Write to: file://{path}")
    if real_run:
        try:
            path.write_text(content, encoding="utf-8")
        except FileNotFoundError:
            path.parent.mkdir(parents=True)
            path.write_text(content, encoding="utf-8")


@dataclasses.dataclass
class WtfMapping:
    """
    定义了一个魔兽世界客户端中被管理的所有 WTF 配置文件的设定.

    :param client: :class:`Client` 对象. 有了这个才知道我们要将配置文件写到哪里去.
    :param all_accounts: 所有的 `wow_acc.api.Account <https://wow-acc.readthedocs.io/en/latest/wow_acc/model.html#wow_acc.model.Account>`_ 对象.
        在 render Jinja 模板时会用到.
    :param all_characters: 所有的 `wow_acc.api.Character <https://wow-acc.readthedocs.io/en/latest/wow_acc/model.html#wow_acc.model.Character>`_ 对象.
        在 render Jinja 模板时会用到.
    :param client_config: :meth:`Client.client_config`.
    :param acc_user_interface: :meth:`Client.get_account_config_cache_wtf`.
    :param acc_macros: :meth:`Client.get_account_macros_cache_txt`.
    :param acc_saved_variables: :meth:`Client.get_account_saved_variables`.
    :param char_user_interface: :meth:`Client.get_character_config_cache_wtf`.
    :param char_chat: :meth:`Client.get_character_chat_cache_txt`.
    :param char_keybinding: :meth:`Client.get_character_bindings_cache_wtf`.
    :param char_layout: :meth:`Client.get_character_layout_local_txt`.
    :param char_addons: :meth:`Client.get_character_addons_txt`.
    :param char_macros: :meth:`Client.get_character_macros_cache_txt`.
    :param char_saved_variables: :meth:`Client.get_character_saved_variables`.
    """

    # fmt: off
    client: Client
    all_accounts: T.Iterable[Account]
    all_characters: T.Iterable[Character]

    client_config: Path = dataclasses.field()

    acc_user_interface: T.List[AccLvlMapping] = dataclasses.field(default_factory=list)
    acc_macros: T.List[AccLvlMapping] = dataclasses.field(default_factory=list)
    acc_saved_variables: T.List[AccLvlMapping] = dataclasses.field(default_factory=list)

    char_user_interface: T.List[CharLvlMapping] = dataclasses.field(default_factory=list)
    char_chat: T.List[CharLvlMapping] = dataclasses.field(default_factory=list)
    char_keybinding: T.List[CharLvlMapping] = dataclasses.field(default_factory=list)
    char_layout: T.List[CharLvlMapping] = dataclasses.field(default_factory=list)
    char_addons: T.List[CharLvlMapping] = dataclasses.field(default_factory=list)
    char_macros: T.List[CharLvlMapping] = dataclasses.field(default_factory=list)
    char_saved_variables: T.List[CharLvlMapping] = dataclasses.field(default_factory=list)
    # fmt: on

    @logger.emoji_block(msg="💻{func_name}", emoji="💻")
    def apply_client_config(self, real_run: bool = False):
        """
        将配置文件写入到 :meth:`Client.client_config`.
        """
        path_out = self.client.client_config
        tpl = Template(self.client_config.read_text())
        content = tpl.render()
        apply(path_out, content, real_run=real_run)

    def _apply_account_level(
        self,
        mapping_list: T.List[AccLvlMapping],
        getter: T.Callable,
        real_run: bool = False,
    ):
        for account, mappings in group_by(
            mapping_list, get_key=lambda x: x.acc
        ).items():
            for mapping in mappings:
                with logger.nested():
                    logger.ruler(f"start 👤account: {account.wtf_account_name!r}")
                    path_out = getter(mapping.acc)
                    content = mapping.tpl.render(
                        account=mapping.acc,
                        all_accounts=self.all_accounts,
                        all_characters=self.all_characters,
                    )
                    apply(path_out, content, real_run=real_run)
                    logger.ruler(f"end 👤account: {account.wtf_account_name!r}")

    @logger.emoji_block(msg="👤🖼{func_name}", emoji="👤")
    def apply_account_user_interface(self, real_run: bool = False):
        """
        将配置文件写入到 :meth:`Client.get_account_config_cache_wtf`.
        """
        self._apply_account_level(
            mapping_list=self.acc_user_interface,
            getter=self.client.get_account_config_cache_wtf,
            real_run=real_run,
        )

    @logger.emoji_block(msg="👤🎮{func_name}", emoji="👤")
    def apply_account_macros(self, real_run: bool = False):  # pragma: no cover
        """
        将配置文件写入到 :meth:`Client.get_account_macros_cache_txt`.
        """
        self._apply_account_level(
            mapping_list=self.acc_macros,
            getter=self.client.get_account_macros_cache_txt,
            real_run=real_run,
        )

    @logger.emoji_block(msg="👤🔢{func_name}", emoji="👤")
    def apply_account_saved_variables(self, real_run: bool = False):
        """
        将配置文件写入到 :meth:`Client.get_account_saved_variables`.
        """
        for mapping in self.acc_saved_variables:
            path_out = self.client.get_account_saved_variables(
                mapping.acc,
                mapping.file.basename,
            )
            content = mapping.tpl.render(
                account=mapping.acc,
                all_accounts=self.all_accounts,
                all_characters=self.all_characters,
            )
            apply(path_out, content, real_run=real_run)

    def _apply_character_level(
        self,
        mapping_list: T.List[CharLvlMapping],
        getter: T.Callable,
        real_run: bool = False,
    ):
        for account, mappings in group_by(
            mapping_list, get_key=lambda x: x.char.account
        ).items():
            for mapping in mappings:
                with logger.nested():
                    logger.ruler(
                        f"start characters in 👤account: {account.wtf_account_name!r}"
                    )
                    path_out = getter(mapping.char)
                    content = mapping.tpl.render(
                        character=mapping.char,
                        all_accounts=self.all_accounts,
                        all_characters=self.all_characters,
                    )
                    apply(path_out, content, real_run=real_run)
                    logger.ruler(
                        f"end characters in 👤account: {account.wtf_account_name!r}"
                    )

    @logger.emoji_block(msg="🧙🖼{func_name}", emoji="🧙")
    def apply_character_user_interface(self, real_run: bool = False):
        """
        将配置文件写入到 :meth:`Client.get_character_config_cache_wtf`.
        """
        self._apply_character_level(
            mapping_list=self.char_user_interface,
            getter=self.client.get_character_config_cache_wtf,
            real_run=real_run,
        )

    @logger.emoji_block(msg="🧙💬{func_name}", emoji="🧙")
    def apply_character_chat(self, real_run: bool = False):
        """
        将配置文件写入到 :meth:`Client.get_character_chat_cache_txt`.
        """
        self._apply_character_level(
            mapping_list=self.char_chat,
            getter=self.client.get_character_chat_cache_txt,
            real_run=real_run,
        )

    @logger.emoji_block(msg="🧙🎹{func_name}", emoji="🧙")
    def apply_character_keybinding(self, real_run: bool = False):
        """
        将配置文件写入到 :meth:`Client.get_character_bindings_cache_wtf`.
        """
        self._apply_character_level(
            mapping_list=self.char_keybinding,
            getter=self.client.get_character_bindings_cache_wtf,
            real_run=real_run,
        )

    @logger.emoji_block(msg="🧙📐{func_name}", emoji="🧙")
    def apply_character_layout(self, real_run: bool = False):
        """
        将配置文件写入到 :meth:`Client.get_character_layout_local_txt`.
        """
        self._apply_character_level(
            mapping_list=self.char_layout,
            getter=self.client.get_character_layout_local_txt,
            real_run=real_run,
        )

    @logger.emoji_block(msg="🧙🧩{func_name}", emoji="🧙")
    def apply_character_addons(self, real_run: bool = False):
        """
        将配置文件写入到 :meth:`Client.get_character_addons_txt`.
        """
        self._apply_character_level(
            mapping_list=self.char_addons,
            getter=self.client.get_character_addons_txt,
            real_run=real_run,
        )

    @logger.emoji_block(msg="🧙🎮{func_name}", emoji="🧙")
    def apply_character_macros(self, real_run: bool = False):  # pragma: no cover
        """
        将配置文件写入到 :meth:`Client.get_character_macros_cache_txt`.
        """
        self._apply_character_level(
            mapping_list=self.char_macros,
            getter=self.client.get_character_macros_cache_txt,
            real_run=real_run,
        )

    @logger.emoji_block(msg="🧙🔢{func_name}", emoji="🧙")
    def apply_character_saved_variables(self, real_run: bool = False):
        """
        将配置文件写入到 :meth:`Client.get_character_saved_variables`.
        """
        for mapping in self.char_saved_variables:
            path_out = self.client.get_character_saved_variables(
                mapping.char,
                mapping.file.basename,
            )
            content = mapping.tpl.render(
                character=mapping.char,
                all_accounts=self.all_accounts,
                all_characters=self.all_characters,
            )
            apply(path_out, content, real_run=real_run)



# -*- coding: utf-8 -*-

from wow_wtf import api


def test():
    _ = api
    _ = api.get_values
    _ = api.group_by
    _ = api.logger
    _ = api.exp03_wotlk

    exp03_wotlk = api.exp03_wotlk
    _ = exp03_wotlk.Client
    _ = exp03_wotlk.AccLvlMapping
    _ = exp03_wotlk.CharLvlMapping
    _ = exp03_wotlk.WtfMapping
    _ = exp03_wotlk.to_module
    _ = exp03_wotlk.Client.get_account_config_cache_wtf
    _ = exp03_wotlk.Client.get_account_bindings_cache_wtf
    _ = exp03_wotlk.Client.get_account_macros_cache_txt
    _ = exp03_wotlk.Client.get_account_saved_variables
    _ = exp03_wotlk.Client.get_character_config_cache_wtf
    _ = exp03_wotlk.Client.get_character_chat_cache_txt
    _ = exp03_wotlk.Client.get_character_bindings_cache_wtf
    _ = exp03_wotlk.Client.get_character_layout_local_txt
    _ = exp03_wotlk.Client.get_character_addons_txt
    _ = exp03_wotlk.Client.get_character_macros_cache_txt
    _ = exp03_wotlk.Client.get_character_saved_variables
    _ = exp03_wotlk.AccLvlMapping.make_many
    _ = exp03_wotlk.CharLvlMapping.make_many
    _ = exp03_wotlk.WtfMapping.apply_client_config
    _ = exp03_wotlk.WtfMapping.apply_account_user_interface
    _ = exp03_wotlk.WtfMapping.apply_account_macros
    _ = exp03_wotlk.WtfMapping.apply_account_saved_variables
    _ = exp03_wotlk.WtfMapping.apply_character_user_interface
    _ = exp03_wotlk.WtfMapping.apply_character_chat
    _ = exp03_wotlk.WtfMapping.apply_character_keybinding
    _ = exp03_wotlk.WtfMapping.apply_character_layout
    _ = exp03_wotlk.WtfMapping.apply_character_addons
    _ = exp03_wotlk.WtfMapping.apply_character_macros
    _ = exp03_wotlk.WtfMapping.apply_character_saved_variables


if __name__ == "__main__":
    from wow_wtf.tests import run_cov_test

    run_cov_test(__file__, "wow_wtf.api", preview=False)



# -*- coding: utf-8 -*-

import typing as T
from ordered_set import OrderedSet


def get_values(enum_class) -> OrderedSet:
    """
    Get all values from an enum class.

    Example::

        >>> class Color:
        ...     RED = 1
        ...     GREEN = 2
        ...     BLUE = 3
        >>> get_values(Color)
        OrderedSet([1, 2, 3])
    """
    return OrderedSet(
        [v for k, v in enum_class.__dict__.items() if not k.startswith("_")]
    )


KT = T.TypeVar("KT")
VT = T.TypeVar("VT")


def group_by(
    iterable: T.Iterable[VT],
    get_key: T.Callable[[VT], KT],
) -> T.Dict[KT, T.List[VT]]:
    """
    Group items by it's key, with type hint.

    Example::

        >>> class Record:
        ...     def __init__(self, product: str, date: str, sale: int):
        ...         self.product = product
        ...         self.date = date
        ...         self.sale = sale

        >>> records = [
        ...     Record("apple", "2020-01-01", 10),
        ...     Record("apple", "2020-01-02", 20),
        ...     Record("apple", "2020-01-03", 30),
        ...     Record("banana", "2020-01-01", 10),
        ...     Record("banana", "2020-01-02", 20),
        ...     Record("banana", "2020-01-03", 30),
        ... ]

        >>> group_by(records, lambda x: x.product)
        {
            "apple": [
                Record("apple", "2020-01-01", 10),
                Record("apple", "2020-01-02", 20),
                Record("apple", "2020-01-03", 30),
            ],
            "banana": [
                Record("banana", "2020-01-01", 10),
                Record("banana", "2020-01-02", 20),
                Record("banana", "2020-01-03", 30),
            ],
        }
    """
    grouped = dict()
    for item in iterable:
        key = get_key(item)
        try:
            grouped[key].append(item)
        except KeyError:
            grouped[key] = [item]
    return grouped



# -*- coding: utf-8 -*-

"""
Wrath of the Lich King (WotLK) expansion module.
"""

from .mapping import Client
from .mapping import AccLvlMapping
from .mapping import CharLvlMapping
from .mapping import WtfMapping
from .dataset import to_module
